/*****************************************************************************/

/*
 *
 *   Copyright (c) 2002, Smart Link Ltd.
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *       1. Redistributions of source code must retain the above copyright
 *          notice, this list of conditions and the following disclaimer.
 *       2. Redistributions in binary form must reproduce the above
 *          copyright notice, this list of conditions and the following
 *          disclaimer in the documentation and/or other materials provided
 *          with the distribution.
 *       3. Neither the name of the Smart Link Ltd. nor the names of its
 *          contributors may be used to endorse or promote products derived
 *          from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 *
 *	modem_debug.c  --  modem debug module (dummy).
 *
 *	Author: Sasha K (sashak@smlink.com)
 *
 *
 */

/*****************************************************************************/

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/bitops.h>
#include <linux/wait.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/errno.h>
#include <asm/uaccess.h>

#include <stdarg.h>

#include <modem.h>
#include <modem_debug.h>

#ifdef CONFIG_DEBUG_MODEM

unsigned modem_debug_level;

/* attributed as weaked symbol in order to provide overriding possibility */
int modem_debug_printf(const char *fmt, ...) __attribute__ ((weak));
int modem_debug_log_data(struct modem *m, unsigned id, const void *data, int count) __attribute__ ((weak));

/* constant definitions */
/* sllog ioctls */
#define SLLOGCTL_SET_DEBUGLEVEL _IOW('s',0x100,unsigned)
#define SLLOGCTL_SET_CATEGORY   _IOW('s',0x101,unsigned)
#define SLLOGCTL_START_MONITOR  _IO('s',0x102)
#define SLLOGCTL_STOP_MONITOR   _IO('s',0x103)

#define SLLOG_PAGE_SHIFT 4


/* data types definitions */
struct sllog_header {
	unsigned modem_id;
	unsigned id;
	unsigned length;
	char     data[0];
};


struct sllog {
	wait_queue_head_t wait;
	unsigned long log_mask;
	unsigned mon;
	/* queue buf */
	int cnt;
	int head;
	int tail;
	int size;
	char *buf;
	char temp[512];
};


static struct sllog *sllog;
static spinlock_t sllog_lock = SPIN_LOCK_UNLOCKED;


/* ---------------------------------------------------------------------- */

static int log_data(struct modem *m, unsigned id, const void *data, int count)
{
	struct sllog *s = sllog;
	struct sllog_header header;
	void *buf;
	unsigned int ret = 0, cnt, header_rest;

	if (!s || !s->mon || id == MODEM_DBG_SKIPME )
		return 0;
	if (id >= sizeof(s->log_mask) * 8) {
		return -1;
	}
	if (!test_bit(id, &s->log_mask))
		return 0;

	header_rest = sizeof(header);
	buf = &header;

	/* check for overflow first */
	if (count + sizeof(header) >= s->size - s->cnt) {
		if (sizeof(header) > s->size - s->cnt)
			goto ret_wakeup;
		header.modem_id = (unsigned)m ;
		//header.modem_id = m ? m->minor : 0;
		header.id = MODEM_DBG_OVERFLOW;
		header.length = 0;
		count = sizeof(header);
	} else {
		header.modem_id = (unsigned)m ;
		//header.modem_id = m ? m->minor : 0;
		header.id = id;
		header.length = count;
		count += sizeof(header);
	}

	while (count) {
		cnt = count;
		if (cnt > s->size - s->head)
			cnt = s->size - s->head;
		if (header_rest) {
			if (cnt > header_rest)
				cnt = header_rest;
			memcpy(s->buf + s->head, buf, cnt);
			header_rest -= cnt;
			buf += cnt;
		} else {
			memcpy(s->buf + s->head, data, cnt);
			data += cnt;
			ret += cnt;
		}
		count -= cnt;
		s->cnt += cnt;
		s->head = (s->head + cnt) % s->size;
	}
  ret_wakeup:
	wake_up_interruptible(&s->wait);
	return ret;
}

int modem_debug_log_data(struct modem *m, unsigned id, const void *data, int count)
{
	unsigned long flags;
	unsigned int ret = 0;

	spin_lock_irqsave(&sllog_lock, flags);
	if(sllog && sllog->mon) {
		ret = log_data(m,id,data,count);
	}
	spin_unlock_irqrestore(&sllog_lock, flags);
	return ret;
}

static int sllog_vprintf(struct modem *m, const char *fmt, va_list args)
{
	struct sllog *s;
	struct timeval tv;
	unsigned long len;
	char *p;
	int count, ret;

	if (!sllog)
		return 0;
	s = sllog;

	if (!fmt || !(len = strlen(fmt)))
		return -1;

	do_gettimeofday(&tv);

	/* skip lead blanks */
	while (*fmt == '\n' || *fmt == '\r' || *fmt == '\t' || *fmt == ' ')
		((char *) fmt)++;

	p = s->temp;
	len = sprintf(p, "<%c%03ld.%06ld> ",
		      in_interrupt()? 'i' : 'p',
		      (tv.tv_sec % 1000), tv.tv_usec);
	p += len;

	ret = vsprintf(p, fmt, args);
	len += ret;

	if(s->temp[len-1] != '\n') {
		s->temp[len++] = '\n';
		s->temp[len]   = '\0';
	}

	if (s->mon) {
		log_data(m, MODEM_DBG_PRINT_MSG, s->temp, len);
		return ret;
	}

	/* store message in sllog buffer */
	count = len;
	ret = 0;
	p = s->temp;

	while (count) {
		int cnt = count;
		if (cnt > s->size - s->cnt)
			cnt = s->size - s->cnt;
		if (cnt > s->size - s->head)
			cnt = s->size - s->head;
		if (cnt<=0)
			break;
		memcpy(s->buf + s->head, p, cnt);
		ret += cnt;
		p += cnt;
		count -= cnt;
		s->head = (s->head + cnt) % s->size;
		s->cnt += cnt;
	}
	/* wake up readers */
	wake_up_interruptible(&s->wait);
	return ret;
}


int modem_debug_printf(const char *fmt, ...)
{
        unsigned long flags;
        int ret = 0;
        spin_lock_irqsave(&sllog_lock,flags);
        if (sllog) {
                va_list args;
                va_start(args, fmt);
                ret = sllog_vprintf(0,fmt,args);
                va_end(args);
        }
        spin_unlock_irqrestore(&sllog_lock,flags);
        return ret;
}

/* ---------------------------------------------------------------------- */


static int sllog_open(struct inode *inode, struct file *file)
{
	struct sllog *s;
	unsigned long flags;

	if (sllog)
		return -EBUSY;

	s = kmalloc(sizeof(*s),GFP_KERNEL);
	if (!s)
		return -ENOMEM;
	memset(s,0,sizeof(*s));
	s->buf = (char *)__get_free_pages(GFP_KERNEL,SLLOG_PAGE_SHIFT);
	if (!s->buf) {
		kfree(s);
		return -ENOMEM;
	}
	init_waitqueue_head(&s->wait);
	s->size = PAGE_SIZE<<SLLOG_PAGE_SHIFT;

	spin_lock_irqsave(&sllog_lock, flags);
	if (sllog) {
		spin_unlock_irqrestore(&sllog_lock, flags);
		kfree(s);
		free_pages((unsigned long)s->buf,SLLOG_PAGE_SHIFT);
		return -EBUSY;
	}
	sllog = s;
	spin_unlock_irqrestore(&sllog_lock, flags);

	file->private_data = s;
	return 0;
}


static int sllog_close(struct inode *inode, struct file *filp)
{
	struct sllog *s = filp->private_data;
	unsigned long flags;
	spin_lock_irqsave(&sllog_lock, flags);
	sllog = 0;
	spin_unlock_irqrestore(&sllog_lock, flags);
	free_pages((unsigned long)s->buf, SLLOG_PAGE_SHIFT);
	kfree(s);
	return 0;
}


static ssize_t sllog_read(struct file *file, char *buffer, size_t count, loff_t * off)
{
	struct sllog *s = file->private_data;
	unsigned long flags;
	int cnt, ret = 0;
	while(count) {
		cnt = count;
		spin_lock_irqsave(&sllog_lock, flags);
		if (cnt > s->cnt)
			cnt = s->cnt;
		if (cnt > s->size - s->tail)
			cnt = s->size - s->tail;
		spin_unlock_irqrestore(&sllog_lock, flags);
		if (cnt <= 0) {
			if (ret) {
				return ret;
			}
			interruptible_sleep_on(&s->wait);
			if (signal_pending(current))
				return -ERESTARTSYS;
			continue;
		}
		if(copy_to_user(buffer, s->buf + s->tail , cnt))
			return -EFAULT;
		ret    += cnt;
		buffer += cnt;
		count  -= cnt;
		spin_lock_irqsave(&sllog_lock, flags);
		s->cnt -= cnt;
		s->tail = (s->tail+cnt)%s->size;
		spin_unlock_irqrestore(&sllog_lock, flags);
	}
	return ret;
}


static int sllog_ioctl(struct inode *pino, struct file *filp, unsigned int cmd,
		       unsigned long arg)
{
	struct sllog *s = filp->private_data;
	unsigned long flags;
	switch (cmd) {
	case SLLOGCTL_SET_DEBUGLEVEL:
		//modem_debug_level = arg;
		return 0;
	case SLLOGCTL_SET_CATEGORY:
		if (arg >= sizeof(s->log_mask) * 8) {
			return -EINVAL;
		}
		set_bit(arg, &s->log_mask);
		return 0;
	case SLLOGCTL_START_MONITOR:
		spin_lock_irqsave(&sllog_lock,flags);
		s->mon = 1;
		s->head = s->tail = s->cnt = 0;
		spin_unlock_irqrestore(&sllog_lock,flags);
		return 0;
	case SLLOGCTL_STOP_MONITOR:
		spin_lock_irqsave(&sllog_lock,flags);
		s->mon = 0;
		s->head = s->tail = s->cnt = 0;
		spin_unlock_irqrestore(&sllog_lock,flags);
		return 0;
	}
	return -ENOIOCTLCMD;
}


static struct file_operations sllog_fops = {
	owner:THIS_MODULE,
	read:sllog_read,
	poll:0,
	ioctl:sllog_ioctl,
	open:sllog_open,
	release:sllog_close
};



/* ---------------------------------------------------------------------- */


static struct proc_dir_entry *proc_modem_dir;
static struct proc_dir_entry* proc_debug;
static struct proc_dir_entry* proc_sllog;

static int proc_read_debug(char *page, char **start, off_t off, int count,
			   int *eof, void *data)
{
        return sprintf(page,"debug = %d\n",modem_debug_level);
}

static int proc_write_debug(struct file *file, const char *buffer,
			    unsigned long count, void *data)
{
	char strval[32];
        int len;
	unsigned n;
	len = count > sizeof(strval) ? sizeof(strval) : count ;
	if(copy_from_user(strval,buffer,len))
		return -EFAULT;
	n = simple_strtoul(strval,NULL,0);
	modem_debug_level = n;
        return len;
}


int modem_debug_init(int debug_level)
{
	spin_lock_init(&sllog_lock);
	modem_debug_level = debug_level;

	proc_modem_dir = proc_mkdir("modem",NULL);
	if(!proc_modem_dir)
		return -1;
	proc_modem_dir->owner = THIS_MODULE;

	proc_debug = create_proc_entry("debug",0644,proc_modem_dir);
	if(!proc_debug) {
		remove_proc_entry("modem", NULL);
		return -1;
	}
	proc_debug->owner = THIS_MODULE;
	proc_debug->read_proc = proc_read_debug;
	proc_debug->write_proc = proc_write_debug;

	proc_sllog = create_proc_entry("sllog",0444,proc_modem_dir);
	if(!proc_sllog) {
		remove_proc_entry("debug",proc_modem_dir);
		remove_proc_entry("modem", NULL);
		return -1;
	}
	proc_sllog->owner = THIS_MODULE;
	proc_sllog->proc_fops  = &sllog_fops;
	return 0;
}

void modem_debug_exit(void)
{
	remove_proc_entry("sllog",proc_modem_dir);
	remove_proc_entry("debug",proc_modem_dir);
	remove_proc_entry("modem", NULL);
}


EXPORT_SYMBOL_NOVERS(modem_debug_level);
EXPORT_SYMBOL_NOVERS(modem_debug_printf);
#endif /* CONFIG_DEBUG_MODEM */
