
/*
 *
 *    Copyright (c) 2002, Smart Link Ltd.
 *    All rights reserved.
 *
 *    Redistribution and use in source and binary forms, with or without
 *    modification, are permitted provided that the following conditions
 *    are met:
 *
 *        1. Redistributions of source code must retain the above copyright
 *           notice, this list of conditions and the following disclaimer.
 *        2. Redistributions in binary form must reproduce the above
 *           copyright notice, this list of conditions and the following
 *           disclaimer in the documentation and/or other materials provided
 *           with the distribution.
 *        3. Neither the name of the Smart Link Ltd. nor the names of its
 *           contributors may be used to endorse or promote products derived
 *           from this software without specific prior written permission.
 *
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *    OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 *
 *    modem_pair.c  --  pair modem device.
 *
 *    Author: Sasha K (sashak@smlink.com)
 *
 *
 */

#include <linux/config.h>
#include <linux/module.h>

#include <linux/init.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/time.h>

#include <modem.h>
#include <modem_debug.h>

#define print_info(fmt...) printk(KERN_INFO fmt)

/* internal data */
static struct pair_state {
	int delay;
	struct timer_list timer;
	int started;
	struct queue {
		int count;
		int head;
		int tail;
		int error;
		int size;
		char buf[4096];
	} a2b, b2a;
	struct io_state {
		struct modem *modem;
		struct io_state *link;
		struct queue *tx;
		struct queue *rx;
		int rings;
		int started;
	} io_a, io_b;
	char buf[1024];
} pair;


#define RING_ON(m)   HZ
#define RING_OFF(m)  5*HZ

static void pair_process(struct io_state *io)
{
	if (!io->started)
		return;
	if (!io->link->started) {
		/* simulate rings */
#if 0
		/* time based rings */
		if(time_before(jiffies,io->link->rings + RING_ON(m)))
			modem_ring(io->link->modem);
		else if(time_after(jiffies,io->link->rings + RING_OFF(m)))
			io->link->rings = jiffies;
#else    
		if(io->link->rings < RING_ON(m))
			modem_ring(io->link->modem);
		else if(io->link->rings > RING_OFF(m))
			io->link->rings = 0;
		io->link->rings++;
#endif
	}
	else {
		struct queue *rx = io->rx, *tx = io->tx;
		int cnt;
	        while(rx->count) {
			cnt = rx->count;
			if(cnt > rx->size - rx->tail)
				cnt = rx->size - rx->tail;
			if(cnt > tx->size - tx->count)
				cnt = tx->size - tx->count;
			if(cnt > tx->size - tx->head)
				cnt = tx->size - tx->head;
			if(cnt<=0)
				break;
			modem_process(io->modem, rx->buf + rx->tail,
				      tx->buf + tx->head, cnt);
			tx->count += cnt;
			tx->head = (tx->head + cnt)%tx->size;
			rx->count -= cnt;
			rx->tail = (rx->tail + cnt)%rx->size;
		}
	}
}

static void pair_interrupt(unsigned long data)
{
	struct pair_state *p = &pair;
	//dprintk("pair_process...\n");
	if(p->started) {
		pair_process(&p->io_a);
		pair_process(&p->io_b);
		pair.timer.expires = jiffies + 1;
		pair.timer.data = 0;
		pair.timer.function = pair_interrupt;
		add_timer(&pair.timer);
	}
}


static int pair_open(struct modem *m)
{
	dprintk("pair_open ...\n");
	return 0;
}

static int pair_close(struct modem *m)
{
	dprintk("pair_close ...\n");
	return 0;
}

static int pair_start(struct modem *m)
{
	struct io_state *io = m->driver_data;
	dprintk("pair_start...\n");
	pair.delay = io->modem->frag;
	memset(io->tx->buf,0,pair.delay);
	io->tx->count = pair.delay;
	io->tx->head  = pair.delay;
	io->started = 1;
	io->rings = 0;
	if (!pair.started) {
		pair.timer.expires = jiffies + 1;
		pair.timer.data = 0;
		pair.timer.function = pair_interrupt;
		add_timer(&pair.timer);
	}
	pair.started++;
	return 0;
}

static int pair_stop(struct modem *m)
{
	struct io_state *io = m->driver_data;
	dprintk("pair_stop ...\n");
	io->started = 0;
	pair.started--;
	if(!pair.started)
		del_timer(&pair.timer);
	return 0;
}

static int pair_ioctl(struct modem *m, unsigned cmd, unsigned long arg)
{
	dprintk("pair_ioctl: cmd %x...\n",cmd);
	switch (cmd) {
	case MDMCTL_CAPABILITIES:
		break;
	case MDMCTL_HOOKSTATE:
		return 0;
	case MDMCTL_CODECTYPE:
		return CODEC_UNKNOWN;
	case MDMCTL_IODELAY:
		return pair.delay;
	case MDMCTL_SPEED:
		return 0;
	case MDMCTL_GETFMTS:
		return MFMT_U16_LE|MFMT_S16_LE;
	case MDMCTL_SETFMT:
		return 0;
	case MDMCTL_SETFRAG:
		return 0;
	default:
		return -ENOIOCTLCMD;
	}
	return 0;
}


/*
 *  device emulation
 *
 */



static struct modem_driver pair_driver = 
{
	name:  "pair driver",
	open:  pair_open,
	close: pair_close,
	start: pair_start,
	stop:  pair_stop,
	ioctl: pair_ioctl,
};


static int __init pair_init(void)
{
	init_timer(&pair.timer);
	pair.a2b.size = sizeof(pair.a2b.buf);
	pair.b2a.size = sizeof(pair.b2a.buf);

	pair.io_a.link = &pair.io_b;
	pair.io_a.tx = &pair.a2b;
	pair.io_a.rx = &pair.b2a;
	pair.io_a.started = 0;
	pair.io_b.link = &pair.io_a;
	pair.io_b.tx = &pair.b2a;
	pair.io_b.rx = &pair.a2b;
	pair.io_b.started = 0;

	pair.io_a.modem = modem_create(&pair_driver,"pair modem dev A");
	if(!pair.io_a.modem) {
		return -ENOMEM;
	}
	pair.io_a.modem->driver_data = &pair.io_a;

	pair.io_b.modem = modem_create(&pair_driver,"pair modem dev B");
	if(!pair.io_b.modem) {
		modem_delete(pair.io_a.modem);
		return -ENOMEM;
	}
	pair.io_b.modem->driver_data = &pair.io_b;

        print_info("modem pair loaded.\n");
	return 0;
}

static void __exit pair_exit(void)
{
	modem_delete(pair.io_a.modem);
	modem_delete(pair.io_b.modem);
}

module_init(pair_init);
module_exit(pair_exit);

MODULE_AUTHOR("sashak");
MODULE_DESCRIPTION("modem pair driver");
MODULE_LICENSE("Smart Link Ltd.");
//MODULE_LICENSE("GPL");

