
/*
 *
 *    Copyright (c) 2002, Smart Link Ltd.
 *    All rights reserved.
 *
 *    Redistribution and use in source and binary forms, with or without
 *    modification, are permitted provided that the following conditions
 *    are met:
 *
 *        1. Redistributions of source code must retain the above copyright
 *           notice, this list of conditions and the following disclaimer.
 *        2. Redistributions in binary form must reproduce the above
 *           copyright notice, this list of conditions and the following
 *           disclaimer in the documentation and/or other materials provided
 *           with the distribution.
 *        3. Neither the name of the Smart Link Ltd. nor the names of its
 *           contributors may be used to endorse or promote products derived
 *           from this software without specific prior written permission.
 *
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *    OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 *
 *    modem.c  --  modem core module.
 *
 *    Author: Sasha K (sashak@smlink.com)
 *
 *
 */

#include <linux/config.h>
#include <linux/module.h>

#include <linux/ctype.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/serial.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>

#include <modem.h>
#include <modem_debug.h>

#define NUM_OF_MODEMS 32

#define MODEM_AUTHOR "Smart Link Ltd."
#define MODEM_NAME "SmartLink Soft Modem"
#define MODEM_VERSION "2.8.3"
#define MODEM_DATE __DATE__" "__TIME__

const char modem_author[] = MODEM_AUTHOR;
const char modem_name[]   = MODEM_NAME;
const char modem_version[]= MODEM_VERSION;
const char modem_date[]   = MODEM_DATE;

#define MODEM_DBG(fmt,arg...) dprintk("%s: " fmt , m->name , ##arg)
#define MODEM_ERR(fmt,arg...) eprintk("%s: " fmt , m->name , ##arg)


/* main modem table */
static struct modem *modem_table[NUM_OF_MODEMS];
static DECLARE_MUTEX(modem_sem);

static char *default_country = NULL;


/* local funcs prototypes */
static struct dp_operations *get_dp_operations(enum DP_ID id);

static int modem_get_chars(struct modem *m, char *buf, int n);
static int modem_put_chars(struct modem *m, const char *buf, int n);
static int modem_comp_get_chars(struct modem *m, char *buf, int n);
static int modem_comp_put_chars(struct modem *m, const char *buf, int n);
static void modem_tty_flush_buffer(struct tty_struct *tty);
int modem_answer(struct modem *m);


/*
 *    misc macros
 */

#define ESCAPE_CHAR(m)        ((m)->sregs[SREG_ESCAPE_CHAR])
#define CR_CHAR(m)            ((m)->sregs[SREG_CR_CHAR])
#define LF_CHAR(m)            ((m)->sregs[SREG_LF_CHAR])
#define BS_CHAR(m)            ((m)->sregs[SREG_BS_CHAR])

#define IS_ECHO(m)            ((m)->sregs[SREG_ECHO])
#define IS_QUIET(m)           ((m)->sregs[SREG_QUIET])
#define IS_VERBOSE(m)         ((m)->sregs[SREG_VERBOSE])
#define IS_AUTOMODE(m)        MODEM_AUTOMODE(m)
#define MODEM_EC_ENABLE(m)    ((m)->sregs[SREG_EC])
#define MODEM_EC_DETECTOR(m)  ((m)->sregs[SREG_EC])
#define MODEM_COMP_ENABLE(m)  ((m)->sregs[SREG_COMP]&&MODEM_EC_ENABLE(m))
#define ESCAPE_TIMEOUT(m)     (HZ/2)
#define ANSWER_DELAY(m)       ((m)->sregs[SREG_ANS_DELAY])
#define SPEAKER_CONTROL(m)    ((m)->sregs[SREG_SPEAKER_CONTROL])
#define SPEAKER_VOLUME(m)     ((m)->sregs[SREG_SPEAKER_VOLUME])

/*
 *    dp operations drivers
 *
 */

struct dp_driver modem_dp_drivers[] = {
        {DP_CID,"CID"},
        {DP_CALLPROG,"CallProg"},
        {DP_DUMMY,"Dummy"},
        {DP_AUTOMODE,"Automode"},
        {DP_V8,"V8"},
        {DP_V17,"V17"},
        {DP_V21,"V21","21"},
        {DP_V22,"V22","22"},
        {DP_V23,"V23","23"},
        {DP_V22BIS,"V22bis","122"},
        {DP_V32,"V32","32"},
        {DP_V32BIS,"V32bis","132"},
        {DP_V34,"V34","34"},
        {DP_B103,"Bell103","103"},
        {DP_B212,"Bell212","212"},
        {DP_FAX,"VFax"},
        {DP_VOICE,"Voice"},
        {DP_K56,"K56Flex","56"},
        {DP_V8BIS,"V8bis"},
        {DP_V90,"V90","90"},
        {DP_V92,"V92","92"},
        {DP_TEST ,"Test"},
        {DP_SINUS,"Sinus"},
        {}
};

static struct dp_driver *get_dp_driver(enum DP_ID id)
{
	struct dp_driver *p;
	for(p = modem_dp_drivers ; p->id > 0 ; p++)
		if (p->id == id) return p;
	return NULL;
}

static struct dp_operations *get_dp_operations(enum DP_ID id)
{
	struct dp_driver *p;
	for(p = modem_dp_drivers ; p->id > 0 ; p++)
		if (p->id == id) return p->op;
	return NULL;
}

int modem_dp_register(enum DP_ID id, struct dp_operations *op)
{
	struct dp_driver *p;
	int ret = -EINVAL;
	down(&modem_sem);
	for(p = modem_dp_drivers ; p->id > 0 ; p++) {
		if (p->id == id && !p->op) {
			p->op = op;
			ret = 0;
			break;
		}
	}
	up(&modem_sem);
	return ret;
}

void modem_dp_deregister(enum DP_ID id, struct dp_operations *op)
{
	struct dp_driver *p;
	down(&modem_sem);
	for(p = modem_dp_drivers ; p->id > 0 ; p++)
		if (p->id == id) p->op = NULL;
	up(&modem_sem);
}



/*
 *    state and status handling
 *
 */

static void modem_report_result(struct modem *m, unsigned code)
{
#define MODEM_RESPONSE_TEXT(id) modem_responses[id].text
#define MODEM_RESPONSE_CODE(id) modem_responses[id].code
	/* response codes : FIXME */
	static const struct modem_response {
		unsigned id;
		char *code;
		char *text;
	} modem_responses[] = {
		{STATUS_OK,         "0", "OK"},
		{STATUS_CONNECT,    "1", "CONNECT"},
		{STATUS_RING,       "2", "RING"},
		{STATUS_NOCARRIER,  "3", "NO CARRIER"},
		{STATUS_ERROR,      "4", "ERROR"},
		{STATUS_INPROGRESS, "" , ""},
		{STATUS_NODIALTONE, "6", "NO DIALTONE"},
		{STATUS_BUSY,       "7", "BUSY"},
		{STATUS_NOANSWER,   "8", "NO ANSWER"},
	};
	static const char none_str[] = "NONE";
        const char *msg;
	u8 msg_mask;

        MODEM_DBG("modem report result: %d (%s)\n",code,MODEM_RESPONSE_TEXT(code));
        if (IS_QUIET(m))
                return;
        if (IS_VERBOSE(m)) {
		if(code==STATUS_CONNECT) {
			msg_mask = modem_get_sreg(m,SREG_CONNNECT_MSG_FORMAT);
			if(msg_mask&1) {
				struct dp_driver *dp_drv;
				modem_put_chars(m,"Modulation: ",12);
				dp_drv = get_dp_driver(m->dp->id);
				msg = (dp_drv) ? dp_drv->name : none_str;
				modem_put_chars(m,msg,strlen(msg));
				modem_put_chars(m,CRLF_CHARS(m),2);
			}
			if(msg_mask&2) {
				modem_put_chars(m,"Protocol: ",10);
				msg = m->cfg.ec ? "LAPM" : none_str;
				modem_put_chars(m,msg,strlen(msg));
				modem_put_chars(m,CRLF_CHARS(m),2);
			}
			if(msg_mask&3) {
				modem_put_chars(m,"Compression: ",13);
				msg = (m->cfg.ec && m->cfg.comp)? "V42bis" : none_str;
				modem_put_chars(m,msg,strlen(msg));
				modem_put_chars(m,CRLF_CHARS(m),2);
			}
			if(msg_mask&4 && m->tx_rate) {
				char rate_str[16];
				modem_put_chars(m,"TxRate: ",8);
				sprintf(rate_str,"%d",m->tx_rate);
				modem_put_chars(m,rate_str,strlen(rate_str));
				modem_put_chars(m,CRLF_CHARS(m),2);
			}
		}
                msg = MODEM_RESPONSE_TEXT(code);
		modem_put_chars(m,msg,strlen(msg));
		if (code==STATUS_CONNECT && m->rx_rate
		    && modem_get_sreg(m,SREG_X_CODE) != 0) {
			char rate_str[16];
			sprintf(rate_str," %d",m->rx_rate);
			modem_put_chars(m,rate_str,strlen(rate_str));
		}
        }
        else {
                msg = MODEM_RESPONSE_CODE(code);
		modem_put_chars(m,msg,strlen(msg));
        }
        modem_put_chars(m,CRLF_CHARS(m),2);
}


/* TDB: state lifecycle description */

/* modem state constants */
#define STATE_MODEM_IDLE      0x1
#define STATE_DP_ESTAB        0x2
#define STATE_EC_ESTAB        0x4
#define STATE_MODEM_ONLINE    0x5
#define STATE_COMMAND_ONLINE  0x6
#define STATE_EC_DISC         0x7
#define STATE_DP_DISC         0x9

#define STATE_DISC 0x100

#define STATE_ESTAB           STATE_DP_ESTAB

#define IS_STATE_LINKED(stat) ((stat) > STATE_DP_ESTAB)

#define IS_STATE_IDLE(stat) ((stat) == STATE_MODEM_IDLE)
#define IS_STATE_CONNECTING(stat) ((stat) > STATE_MODEM_IDLE && (stat) < STATE_MODEM_ONLINE)
#define IS_STATE_ONLINE(stat) ((stat) == STATE_MODEM_ONLINE || (stat) == STATE_COMMAND_ONLINE)
#define IS_STATE_DISC(stat) ((stat) == STATE_EC_DISC || (stat) == STATE_DP_DISC)

#define IS_STATE_COMMAND(stat) ((stat)==STATE_MODEM_IDLE || (stat)==STATE_COMMAND_ONLINE)

static int modem_set_state(struct modem *m, unsigned new_state)
{
        unsigned long flags;
        MODEM_DBG("modem set state: %x --> %x...\n",
		  m->state, new_state);
	if(m->state == new_state)
		return 0;
        spin_lock_irqsave(&m->lock, flags);
	switch(new_state) {
	case STATE_MODEM_IDLE:
		MODEM_DBG("new state: MODEM_IDLE\n");
		m->get_chars = 0;
		m->put_chars = 0;
		m->get_bits  = 0;
		m->put_bits  = 0;
		break;
	case STATE_DP_ESTAB:
		MODEM_DBG("new state: DP_ESTAB\n");
		m->get_chars = 0;
		m->put_chars = 0;
		m->get_bits  = 0;
		m->put_bits  = 0;
		break;
	case STATE_EC_ESTAB:
		MODEM_DBG("new state: EC_ESTAB\n");
		m->get_chars = 0;
		m->put_chars = 0;
		break;
	case STATE_MODEM_ONLINE:
		MODEM_DBG("new state: MODEM_ONLINE\n");
		m->put_chars = (m->cfg.ec && m->cfg.comp&0x1)
			? modem_comp_put_chars : modem_put_chars;
		m->get_chars = (m->cfg.ec && m->cfg.comp&0x2)
			? modem_comp_get_chars : modem_get_chars;
		m->modem_info |= TIOCM_CD;
		break;
	case STATE_COMMAND_ONLINE:
		MODEM_DBG("new state: COMMAND_ONLINE\n");
		m->get_chars = 0;
		m->put_chars = 0;
		break;
	case STATE_EC_DISC:
		MODEM_DBG("new state: EC_DISC\n");
		m->get_chars = 0;
		m->put_chars = 0;
		break;
	case STATE_DP_DISC:
		MODEM_DBG("new state: DP_DISC\n");
		m->get_chars = 0;
		m->put_chars = 0;
		m->get_bits  = 0;
		m->put_bits  = 0;
		break;
	default:
		MODEM_ERR("invalid state: %d\n", new_state);
		return -EINVAL;
	}
        m->state = new_state;
        spin_unlock_irqrestore(&m->lock, flags);
	return 0;
}


static void modem_hup(struct modem *m, unsigned local)
{
	MODEM_DBG("modem_hup...\n");
	switch(m->state) {
	case STATE_MODEM_IDLE:
		return;
	case STATE_DP_ESTAB:
		if(local && m->dp && m->dp->op->hangup )
			m->dp->op->hangup(m->dp);
		modem_set_state(m,STATE_DP_DISC);
		schedule_task(&m->stop_task);
		break;
	case STATE_EC_ESTAB:
	case STATE_MODEM_ONLINE:
	case STATE_COMMAND_ONLINE:
		if(local && m->cfg.ec) {
			modem_ec_stop(m);
			modem_set_state(m,STATE_EC_DISC);
		}
		else {
			if(local && m->dp && m->dp->op->hangup)
				m->dp->op->hangup(m->dp);
			modem_set_state(m,STATE_DP_DISC);
		}
		schedule_task(&m->stop_task);
		if(m->modem_info&TIOCM_CD && m->tty && !C_CLOCAL(m->tty)) {
			MODEM_DBG("tty_hangup...\n");
			tty_hangup(m->tty);
		}
		m->modem_info &= ~TIOCM_CD;
		break;
	case STATE_EC_DISC:
		if(local && m->dp && m->dp->op->hangup )
			m->dp->op->hangup(m->dp);
		modem_set_state(m,STATE_DP_DISC);
		break;
	case STATE_DP_DISC:
	default:
		return;
	}
}


void modem_update_status(struct modem *m, unsigned status)
{
	MODEM_DBG("modem_update_status: %d\n", status);
	switch(status) {
	case STATUS_OK:
		break;
	case STATUS_DP_LINK:
		MODEM_DBG("--> DP LINK\n");
		modem_set_state(m,STATE_EC_ESTAB);
		/* start packer */
		if(m->cfg.ec && m->cfg.ec_detector) {
			modem_detector_start(m);
			m->get_bits = modem_detector_get_bits;
			m->put_bits = modem_detector_put_bits;
		}
		else {
			/* pass pack connect phase */
			modem_update_status(m,STATUS_PACK_LINK);
		}
		break;
	case STATUS_PACK_LINK:
		MODEM_DBG("--> PACK LINK\n");
		if(m->cfg.ec) {
			/* start EC */
			modem_hdlc_start(m);
			m->get_bits = modem_hdlc_get_bits;
			m->put_bits = modem_hdlc_put_bits;
			modem_ec_start(m);
		}
		else {
			/* start async, pass EC connect phase */
			modem_async_start(m);
			m->get_bits = modem_async_get_bits;
			m->put_bits = modem_async_put_bits;
			modem_update_status(m,STATUS_EC_LINK);
		}
		break;
	case STATUS_CONNECT:
	case STATUS_EC_LINK:
		MODEM_DBG("--> EC LINK\n");
		modem_set_state(m,STATE_MODEM_ONLINE);
		modem_report_result(m,STATUS_CONNECT); // fixme
		if(SPEAKER_CONTROL(m) == 1)
			m->driver.ioctl(m,MDMCTL_SPEAKERVOL,0);
		break;
	case STATUS_EC_RELEASE:
	case STATUS_EC_ERROR:
		MODEM_DBG("--> EC UNLINK\n");
		modem_hup(m,(m->state == STATE_EC_DISC));
		break;
	default:
		MODEM_DBG("unknown status.\n");
	case STATUS_ERROR:
	case STATUS_DP_ERROR:
		status = STATUS_NOCARRIER;
	case STATUS_NOCARRIER:
	case STATUS_NODIALTONE:
	case STATUS_BUSY:
	case STATUS_NOANSWER:
		MODEM_DBG("--> FINISH.\n");
		m->result_code = status;
		modem_hup(m,IS_STATE_DISC(m->state));
		break;
	}
}



void modem_update_config(struct modem *m, struct modem_config *cfg)
{
	MODEM_DBG("modem update config...\n");
#if 1
#define PRINT_CONFIG(prm) MODEM_DBG(#prm " = %d (%d)\n",cfg->prm,m->cfg.prm)
	PRINT_CONFIG(ec);
	PRINT_CONFIG(ec_detector);
	PRINT_CONFIG(ec_tx_win_size);
	PRINT_CONFIG(ec_rx_win_size);
	PRINT_CONFIG(ec_tx_info_size);
	PRINT_CONFIG(ec_rx_info_size);
	PRINT_CONFIG(comp);
	PRINT_CONFIG(comp_dict_size);
	PRINT_CONFIG(comp_max_string);
#endif
	// FIXME, FIXME, FIXME !!!
#if 0	
	if(cfg->ec) ; /* TBD: now ec reconfig is handled internally in EC */
#endif
	if(cfg->comp && m->cfg.comp) {
		m->cfg.comp = cfg->comp;
		if(!modem_comp_config(m,cfg->comp_dict_size,
				      cfg->comp_max_string)) {
			m->cfg.comp_dict_size  = cfg->comp_dict_size;
			m->cfg.comp_max_string = cfg->comp_max_string;
		}
		else { /* config failed */
			; /* TBD: probably try to renegotiate */
		}
	}
	else
		m->cfg.comp = 0;
}


/*
 *    Interrupt handlers and services
 *
 */

/* get/put bits */
int modem_get_bits(struct modem *m, int nbits, u8 *buf, int n)
{
	if(m->get_bits)
		n = m->get_bits(m,nbits,buf,n);
	else
		memset(buf,0xff,n);
	modem_debug_log_data(m,MODEM_DBG_TX_BITS,buf,n);
	/* process bit timer and packer */
	if (m->bit_timer && (m->bit_timer -= nbits*n) <= 0) {
		m->bit_timer = 0;
		m->bit_timer_func(m);
	}
	// FIXME NOW: remove packer_process from modem to EC
	if(m->packer_process)
		m->packer_process(m,nbits*n);
	return n;
}

int modem_put_bits(struct modem *m, int nbits, u8 *buf, int n)
{
	if(m->put_bits)
		n = m->put_bits(m,nbits,buf,n);
	modem_debug_log_data(m,MODEM_DBG_RX_BITS,buf,n);	
	return n;
}


// fixme: 
extern void *dcr_create();
extern void  dcr_delete(void *dcr);
extern void  dcr_process(void *dcr, void *buf, int len);

static void modem_dp_process(struct modem *m,void *in,void *out,int count)
{
	struct dp *dp;
	int ret;
	unsigned cnt;
	//MODEM_DBG("modem_process %d: %d...\n", m->count,count);
	while(count) {
		cnt = count;
		if (cnt > m->frag)
			cnt = m->frag;
		dp = m->dp;
		if (!dp) {
			m->process = NULL;
			memset(out,0,count);
			return;
		}
		while(dp) {
			//MODEM_DBG("%d: running dp %s...\n", m->count,dp->name);
			ret = dp->op->process(m->dp,in,out,cnt>>MFMT_SHIFT(m->format));
			switch(ret) {
			case STATUS_OK:
				break;
			case STATUS_CONNECT:
				if(!IS_STATE_LINKED(m->state) &&
				   !IS_STATE_DISC(m->state))
					modem_update_status(m,STATUS_DP_LINK);
				break;
			case STATUS_CHANGEDP:
				schedule_task(&m->change_dp_task);
				break;
			default: // errors
				modem_update_status(m,ret);
				m->process = NULL;
				memset(out,0,count);
				return;
			}
			dp = dp->next;
		}
		count -= cnt;
		out += cnt;
		in  += cnt;
	}
}

void modem_process(struct modem *m,void *in,void *out,int count)
{
	/* clean DC */
	dcr_process(m->dcr,in,count>>MFMT_SHIFT(m->format));
	modem_debug_log_data(m,MODEM_DBG_RX_SAMPLES,in,count);
	if(m->process)
		m->process(m,in,out,count);
	else {
		/* mute output */
		memset(out,0,count);
		if ( m->expires && m->count >= m->expires ) {
			m->expires = 0;
			schedule_task(&m->change_dp_task);
		}
	}
	modem_debug_log_data(m,MODEM_DBG_TX_SAMPLES,out,count);
	m->count+=count>>MFMT_SHIFT(m->format); // fixme
}


void modem_error(struct modem *m)
{
	MODEM_DBG("modem_error...\n");
}


/* ring detection */
#define RING_ON_MIN(m)   HZ*3/20 /* 0.15 sec */
#define RING_ON_MAX(m)   HZ*5/2  /* 2.5 sec */
#define RING_OFF_MIN(m)  HZ*3/2  /* 1.5 sec */
#define RING_OFF_MAX(m)  HZ*13/2 /* 6.5 sec */
#define RING_COUNT_MIN(m) 10

#define TOTAL_RINGS_COUNT(m) (m)->sregs[SREG_RING_COUNTER]
#define ANSWER_AFTER_RINGS(m) (m)->sregs[SREG_RINGS_TO_AUTO_ANSWER]

static void ring_cancel(unsigned long data)
{
	struct modem *m = (struct modem *)data;
	MODEM_DBG("ring_cancel...\n");
	TOTAL_RINGS_COUNT(m) = 0;
	m->ring_count = 0;
}

static void ring_finish(unsigned long data)
{
	struct modem *m = (struct modem *)data;
	MODEM_DBG("ring_finish...\n");
	if ( time_before(m->ring_last,m->ring_first + RING_ON_MIN(m)) || // too short
	     time_after(m->ring_last,m->ring_first + RING_ON_MAX(m))  || // too long
	     m->ring_count < RING_COUNT_MIN(m)) {        // low freq
		MODEM_DBG("ring reject: count %d (%lu-%lu)\n",
			  m->ring_count,m->ring_first,m->ring_last);
		ring_cancel(data);
		return;
	}
	m->ring_count = 0;
	if(!TOTAL_RINGS_COUNT(m)++)
		modem_put_chars(m,CRLF_CHARS(m),2);
	modem_report_result(m, STATUS_RING);
	/* check for CID,answer */
#if CONFIG_CID
	if (TOTAL_RINGS_COUNT(m)) {
		/* TODO: start CID */;
	}
#endif
	if (ANSWER_AFTER_RINGS(m) &&
	    TOTAL_RINGS_COUNT(m) >= ANSWER_AFTER_RINGS(m)) {
		TOTAL_RINGS_COUNT(m) = 0;
		modem_answer(m);
	}
	else {
		m->timer.data     = (unsigned long)m;
		m->timer.function = ring_cancel;
		m->timer.expires = jiffies + RING_OFF_MAX(m) + 1;
		add_timer(&m->timer);
	}
}

void modem_ring(struct modem *m)
{
	//MODEM_DBG("modem_ring...\n");
	if(m->state != STATE_MODEM_IDLE)
		return;
	if(!m->ring_count) {
		if(time_before(jiffies,m->ring_last + RING_OFF_MIN(m))) {
			MODEM_DBG("bad ring: now %lu, last %lu.\n",
				  jiffies,m->ring_last);
			m->ring_last = jiffies;
			return;
		}
		del_timer(&m->timer);
		m->timer.data     = (unsigned long)m;
		m->timer.function = ring_finish;
		m->timer.expires = jiffies + RING_ON_MAX(m) + 1;
		add_timer(&m->timer);
		m->ring_first = jiffies;
		MODEM_DBG("modem_ring: timer +%d\n",RING_ON_MAX(m) + 1);
	}
	m->ring_last = jiffies;
	m->ring_count++;
}


/* command mode processing */
static void modem_at_process(void *data)
{
	extern int process_at_command(struct modem *m, char *buf);
	struct modem *m = (struct modem *)data;
	int echo = IS_ECHO(m);
	char lf = LF_CHAR(m), cr = CR_CHAR(m), bs = BS_CHAR(m);
        int ret;
	char ch;
	if(!IS_STATE_COMMAND(m->state))
		return;
	while(1) {
		if(modem_get_chars(m,&ch,1) <= 0)
			break;
		if (ch == '/' && m->at_count == 1 &&
		    toupper(m->at_line[0]) == 'A') {
			m->at_count = strlen(m->at_line);
			if(echo)
				modem_put_chars(m,m->at_line+1,m->at_count-1);
			ch = cr;
		}
		if (echo)
			modem_put_chars(m,&ch,1);
		if (ch == cr || ch == lf) {
			int i;
			char *p = m->at_cmd;
			m->at_line[m->at_count] = '\0';
			for(i = 0 ; i < m->at_count ; i++) {
				if( m->at_line[i] == ' ' ||
				    m->at_line[i] == '\t' )
					continue;
				*p++ = m->at_line[i];
			}
			*p = '\0';
			if (strlen(m->at_cmd) < 2 ||
			    toupper(m->at_cmd[0]) != 'A' ||
			    toupper(m->at_cmd[1]) != 'T' ) {
				memset(m->at_line,' ',m->at_count);
				if(echo) {
					modem_put_chars(m,&cr,1);
					modem_put_chars(m,m->at_line,m->at_count);
					modem_put_chars(m,&cr,1);
				}
				m->at_count = 0;
				continue;
			}
			if (echo)
				modem_put_chars(m,CRLF_CHARS(m),2);
			MODEM_DBG("run cmd: %s\n",m->at_cmd);
			ret = process_at_command(m, m->at_cmd);
			if (ret < 0)
				modem_report_result(m,STATUS_ERROR);
			else if (ret==0)
				modem_report_result(m,STATUS_OK);
			m->at_line[m->at_count] = '\0';
			m->at_count = 0;
			echo = IS_ECHO(m);
			cr = CR_CHAR(m), lf = LF_CHAR(m), bs = BS_CHAR(m);
		}
		else if (ch == bs && m->at_count) {
			m->at_count--;
			if(echo) {
				ch = ' ';
				modem_put_chars(m,&ch,1);
				modem_put_chars(m,&bs,1);
			}
		}
		else if (m->at_count == sizeof(m->at_line) - 1) {
			if (echo)
				modem_put_chars(m,CRLF_CHARS(m),2);
			m->at_count   = 0;
			m->at_line[1] = '\0';
			modem_report_result(m,STATUS_ERROR);
		}
		else {
			m->at_line[m->at_count++] = ch;
		}
	}
}


/*
 *    TTY procedures
 *
 */


/* get chars */
static int modem_get_chars(struct modem *m, char *buf, int n)
{
	unsigned long flags;
	int ret = 0, cnt;
	spin_lock_irqsave(&m->lock,flags);
        while(n) {
                cnt = n;
                if (cnt > m->xmit.count)
                        cnt = m->xmit.count;
                if (cnt > m->xmit.size - m->xmit.tail)
                        cnt = m->xmit.size - m->xmit.tail;
                if (cnt <= 0) {
                        break;
                }
                memcpy(buf, m->xmit.buf + m->xmit.tail, cnt);
                ret += cnt;
                n -= cnt;
                buf += cnt;
                m->xmit.count -= cnt;
                m->xmit.tail = (m->xmit.tail+cnt)%m->xmit.size;
        }
	if(ret && m->tty) {
		struct tty_struct *tty = m->tty;
		if ((test_bit(TTY_DO_WRITE_WAKEUP, &tty->flags)) &&
		    tty->ldisc.write_wakeup)
			(tty->ldisc.write_wakeup)(tty);
		wake_up_interruptible(&tty->write_wait);
	}
	spin_unlock_irqrestore(&m->lock,flags);
	//MODEM_DBG("modem_get_chars: %d...\n", n);
	return ret;
}

/* put chars */
static int modem_put_chars(struct modem *m, const char *buf, int n)
{
	unsigned long flags;
	int i;
        if(!m->tty)
                return 0;
        spin_lock_irqsave(&m->lock,flags);
        for ( i = 0 ; i < n && m->tty->flip.count < TTY_FLIPBUF_SIZE ; i++ ) {
                tty_insert_flip_char(m->tty,buf[i],0);
        }
        if (m->tty->flip.count > 0)
                tty_flip_buffer_push(m->tty);
        spin_unlock_irqrestore(&m->lock,flags);
	if ( i < n ) {
		MODEM_DBG("modem put_chars (%d): flip is full!!!\n",n-i);
	}
#if 0
	if(i>0) {
		//MODEM_DBG("modem_comp_put_chars: %d...\n",i);
		modem_debug_log_data(m,	MODEM_DBG_ECHOC_SAMPLES,buf,i);
	}
#endif
	return i;
}

/* compressor get/put chars */
/* fixme: unify interfaces: modem tx -> comp -> ec -> hdlc -->
                            modem tx -> ec -> hdlc         -->
                            modem tx -> async              -->
			    modem tx --> raw output
*/
static int modem_comp_get_chars(struct modem *m, char *buf, int n)
{
	int ret = 0, cnt;
	char ch;
	while(ret < n) {
		cnt = modem_get_chars(m,&ch,1);
		if(cnt <= 0)
			break;
		cnt = modem_comp_encode(m,ch,buf+ret,n-ret);
		if(cnt < 0) {
			break;
		}
		ret += cnt;
	}
	if(ret < n) {
		cnt =  modem_comp_flush_encoder(m,buf+ret,n-ret);
		ret += cnt;
	}
	if(ret > 0) {
		//MODEM_DBG("modem_comp_get_chars: %d(%d)...\n",ret,count);
		modem_debug_log_data(m,MODEM_DBG_TX_DATA,buf,ret);
	}
	return ret;
}

static int comp_send_output(struct modem *m)
{
	int cnt;
	if(m->comp.count > 0) {
		cnt = modem_put_chars(m,m->comp.buf+m->comp.head,m->comp.count);
		m->comp.count -= cnt;
		m->comp.head += cnt;
		if(m->comp.count > 0)
			return 0;
		else
			m->comp.head = 0;
	}
	return 1;
}

static int modem_comp_put_chars(struct modem *m, const char *buf, int n)
{
	int i=0, cnt;
	if(!comp_send_output(m))
		return 0;
	while(i<n) {
		cnt = modem_comp_decode(m,buf[i++],m->comp.buf,sizeof(m->comp.buf));
		if(cnt < 0) {
			MODEM_DBG("decoder error. %d(%d)\n",i,n);
			modem_update_status(m,STATUS_ERROR);
			break;
		}
		m->comp.count += cnt;
		if(!comp_send_output(m))
			break;
	}
	if(i==n) {
		cnt = modem_comp_flush_decoder(m,m->comp.buf,sizeof(m->comp.buf));
		m->comp.count += cnt;
		comp_send_output(m);
	}
	if(i>0) {
		//MODEM_DBG("modem_comp_put_chars: %d...\n",i);
		modem_debug_log_data(m,MODEM_DBG_RX_DATA,buf,i);
	}
	return i;
}




/*
 *    internal procedures
 *
 */

#define IS_FAST_DP(id) (id == DP_V34 || id == DP_V90 || id == DP_V92)

static void do_modem_change_dp (void *data)
{
	struct modem *m = data;
	struct dp_operations *op;
        struct dp *dp = NULL, *old;
	int dp_id,id;
        unsigned long flags;

	spin_lock_irqsave(&m->lock,flags);
	dp_id = m->dp_requested;
	m->dp_requested = 0;
	spin_unlock_irqrestore(&m->lock,flags);

	id = dp_id;
	if(dp_id == 0) {
		dp_id = MODEM_DP(m);
		id = IS_FAST_DP(dp_id) ? DP_V8 : dp_id;
	}

        MODEM_DBG("%d: change dp: --> %d...\n", m->count, id);

	op = get_dp_operations(id);
	if (op && op->create)
		dp = op->create(m,dp_id,
				m->caller,m->srate,
				m->frag>>MFMT_SHIFT(m->format),op);
	if (!dp) {
		MODEM_ERR("change dp -> %d error.\n", id);
		modem_hup(m,1);
	}

        spin_lock_irqsave(&m->lock, flags);
        old = m->dp;
        dp->next = old ? old->next:NULL;
        m->dp = dp;
	m->process = modem_dp_process;
        spin_unlock_irqrestore(&m->lock, flags);
        if ( old ) {
                old->op->delete(old);
        }

}


static int modem_set_hook(struct modem *m, unsigned hook_state)
{
        int ret;
        MODEM_DBG("modem set hook: %d --> %d...\n", m->hook, hook_state);
        if ( m->hook == hook_state )
                return 0;
        ret = m->driver.ioctl(m, MDMCTL_HOOKSTATE,hook_state);
        if (!ret)
                m->hook = hook_state;
        return ret;
}

static void modem_setup_config(struct modem *m)
{
	m->cfg.ec = MODEM_EC_ENABLE(m);
	m->cfg.ec_detector = MODEM_EC_DETECTOR(m);
	m->cfg.ec_tx_win_size  = LAPM_MAX_WIN_SIZE;
	m->cfg.ec_rx_win_size  = LAPM_MAX_WIN_SIZE;
	m->cfg.ec_tx_info_size = LAPM_MAX_INFO_SIZE;
	m->cfg.ec_rx_info_size = LAPM_MAX_INFO_SIZE;
	if(m->cfg.ec && MODEM_COMP_ENABLE(m))
		m->cfg.comp = 0x3; /* bi-directional v42bis */
	else
		m->cfg.comp = 0;
	m->cfg.comp_dict_size  = COMP_MAX_CODEWORDS;
	m->cfg.comp_max_string = COMP_MAX_STRING;
}


static void do_modem_start (void *data)
{
	struct modem *m = data;
        int ret;
        MODEM_DBG("do_modem_start..\n");
	m->result_code = 0;
	modem_setup_config(m);
        modem_set_state(m, STATE_ESTAB);
        ret = m->driver.ioctl(m, MDMCTL_SPEED, m->srate);
        ret = m->driver.ioctl(m, MDMCTL_SETFRAG,m->frag);
	if(SPEAKER_CONTROL(m) == 1)
		m->driver.ioctl(m,MDMCTL_SPEAKERVOL,SPEAKER_VOLUME(m));
        ret = modem_set_hook(m, MODEM_HOOK_OFF);
        if (ret)
                goto error;

	m->xmit.head = m->xmit.tail = m->xmit.count = 0;
	m->dcr = dcr_create();
	if(!m->dcr) {
		ret = -ENOMEM;
		goto error;
	}
	if (m->cfg.ec && m->cfg.comp) {
		ret = modem_comp_init(m);
		if (ret) goto error1;
	}
	if (m->cfg.ec) {
		ret = modem_ec_init(m);
		if (ret) goto error2;
	}

        m->count = 0;
	m->process = NULL;

        ret = m->driver.start(m);
	if (!ret)
		return;

	MODEM_ERR("modem start = %d: cannot start device.\n",ret);

	modem_ec_exit(m);
 error2:
	modem_comp_exit(m);
 error1:
	dcr_delete(m->dcr);
 error:
	MODEM_ERR("modem start: failed.\n");
        modem_set_state(m, STATE_MODEM_IDLE);
	modem_report_result(m, STATUS_NOCARRIER);
}


static void do_modem_stop (void *data)
{
	struct modem *m = (struct modem *) data;
	unsigned long flags;
        int ret = 0;

	set_current_state(TASK_UNINTERRUPTIBLE);
	schedule_timeout(HZ/2);

        MODEM_DBG("do_modem_stop..\n");
        ret = m->driver.stop(m);
        if ( ret ) {
                MODEM_ERR("modem stop = %d: cannot stop device.\n",ret);
        }
        modem_set_hook(m, MODEM_HOOK_ON);
	if(SPEAKER_CONTROL(m) == 1)
		m->driver.ioctl(m,MDMCTL_SPEAKERVOL,0);
	m->caller = 0;

	if(m->dcr)
		dcr_delete(m->dcr);
	// FIXME: If ec,comp were allocated handled inside _exit() - improve?
	modem_ec_exit(m);
	modem_comp_exit(m);

        if (m->dp) {
                struct dp *dp;
                spin_lock_irqsave(&m->lock,flags);
                dp = m->dp;
                m->dp = 0;
                spin_unlock_irqrestore(&m->lock,flags);
                if(dp) dp->op->delete(dp);
        }
        modem_set_state(m, STATE_MODEM_IDLE);
	modem_report_result(m,(m->result_code)?m->result_code:STATUS_NOCARRIER);
	m->result_code = 0;
}


/*
 *    Commands
 *
 */

int modem_send_to_tty(struct modem *m, const char *buf, int n)
{
	return modem_put_chars(m,buf,n);
}

int modem_answer(struct modem *m)
{
        MODEM_DBG("modem answer...\n");
        if ( m->dp ) {
                MODEM_ERR("dp %d is already exists.\n", m->dp->id);
                return -1;
        }
	//m->automode_requested = MODEM_AUTOMODE(m);
	m->dp_requested = 0;
	m->automode_requested = 0;
        m->expires = ANSWER_DELAY(m) ? ANSWER_DELAY(m)*m->srate/100:m->srate/100;
	schedule_task(&m->start_task);
        return 0;
}


int modem_dial(struct modem *m)
{
        MODEM_DBG("modem dial: %s...\n", m->dial_string);
#if 1  // FIXME: temporary for testing
	if(m->dial_string[0]=='c' || m->dial_string[0]=='C') {
		m->caller = 1;
		return modem_answer(m);
	}
#endif
	m->dp_requested = DP_CALLPROG;
	//m->automode_requested = MODEM_AUTOMODE(m);
	m->automode_requested = 0;
        m->expires = m->srate/100;
        m->caller = 1;
	schedule_task(&m->start_task);
	return 0;
}


int modem_hook(struct modem *m, unsigned hook_state)
{
	MODEM_DBG("modem hook...\n");
        if ( m->hook == hook_state )
                return 0;
        if (!IS_STATE_IDLE(m->state))
		modem_hup(m,1);
        return modem_set_hook(m,hook_state);
}

int modem_online(struct modem *m)
{
	MODEM_DBG("modem online...\n");
        if (m->state != STATE_COMMAND_ONLINE)
                return -1;
        modem_set_state(m, STATE_MODEM_ONLINE);
	modem_report_result(m,STATUS_CONNECT); // fixme
        return 0;
}

void modem_update_speaker(struct modem *m)
{
	//MODEM_DBG("modem update speaker...\n");
	m->driver.ioctl(m,MDMCTL_SPEAKERVOL,
			SPEAKER_CONTROL(m) == 2 ? SPEAKER_VOLUME(m):0);
}

int modem_get_sreg(struct modem *m, unsigned int num)
{
	if (num >= sizeof(m->sregs))
		return -1;
	return m->sregs[num];
}

int modem_set_sreg(struct modem *m, unsigned int num, int val)
{
	if (num >= sizeof(m->sregs))
		return -1;
	// fixme: lock
	m->sregs[num] = val;
	return 0;
}

/* homolog set initialization */
int modem_homolog_init(struct modem *m, int id, const char *name)
{
	const struct homolog_set *set;
	for(set=homolog_set;set->name;set++) {
		if(set->id == id ||
		   (name && *name && !strcmp(name,set->name))) {
			m->homolog = set;
			modem_set_sreg(m,SREG_DIAL_PAUSE_TIME,
				       set->params->DialPauseTime);
			modem_set_sreg(m,SREG_FLASH_TIMER,
				       set->params->HookFlashTime);
			return 0;
		}
	}
	return -1;
}

int modem_reset(struct modem *m)
{
	static int sregs_init(unsigned char sregs[]);
	MODEM_DBG("modem reset...\n");
	// TBD
	if(m->state != STATE_MODEM_IDLE) {
		modem_hup(m,1);
	}
	if(m->hook)
		modem_set_hook(m,MODEM_HOOK_ON);
	modem_set_state(m, STATE_MODEM_IDLE);
	sregs_init(m->sregs);
	modem_homolog_init(m,m->homolog->id,NULL);
	return 0;
}



/*
 *    Init functions
 *
 */

/* set default init values */
static int sregs_init(unsigned char sregs[])
{
        sregs[SREG_ESCAPE_CHAR]                  = '+' ; /* escape char */
        sregs[SREG_CR_CHAR]                      = '\r'; /* cr char */
        sregs[SREG_LF_CHAR]                      = '\n'; /* lf char */
        sregs[SREG_BS_CHAR]                      = '\b'; /* bs char */
        sregs[SREG_DIAL_TONE_WAIT_TIME]          =    2; /* seconds */
        sregs[SREG_WAIT_CARRIER_AFTER_DIAL]      =   60; /* seconds */
        sregs[SREG_DIAL_PAUSE_TIME]              =    2; /* seconds */
        sregs[SREG_CARRIER_DETECT_RESPONSE_TIME] =    6; /* 0.1 sec */
        sregs[SREG_CARRIER_LOSS_DISCONNECT_TIME] =    7; /* 0.1 sec */
        sregs[SREG_DTMF_DURATION]                =  100; /* ms */
        sregs[SREG_ESCAPE_PROMPT_DELAY]          =   50; /* ms */
        sregs[SREG_FLASH_TIMER]                  =   20; /* 10ms */
        sregs[SREG_ECHO]                         =    1; /* yes */
        sregs[SREG_QUIET]                        =    0; /* no  */
        sregs[SREG_VERBOSE]                      =    1; /* yes */
        sregs[SREG_TONE_OR_PULSE]                =    1; /* tone */

        sregs[SREG_X_CODE]                       =    4;

        sregs[SREG_SPEAKER_CONTROL] =  1; /* yes */
        sregs[SREG_SPEAKER_VOLUME]  =  3; /* max */
        sregs[SREG_AUTOMODE]        =  1; /* yes */
        sregs[SREG_DP]              =  DP_V90;

        sregs[SREG_ANS_DELAY]       =  2; /* seconds */

        sregs[SREG_LINE_QUALITY_CONTROL]   = 0;
        sregs[SREG_CD]                     = 0;
        sregs[SREG_FLOW_CONTROL]           = 0;
        sregs[SREG_CONNNECT_MSG_FORMAT]    = 0;
        sregs[SREG_CONNNECT_MSG_SPEED_SRC] = 0;

	/* new sregs */
	sregs[SREG_EC]   = 1;
	sregs[SREG_COMP] = 0x3;

        return 0;
}


/*
 *   tty driver
 *
 */

static int modem_tty_write_room(struct tty_struct *tty)
{
	struct modem *m = tty->driver_data;
	//MODEM_DBG("modem tty write room: %d...\n", m->xmit.size - m->xmit.count);
	return m->xmit.size - m->xmit.count;
}

#if 0
static void modem_tty_send_xchar(struct tty_struct *tty, char ch)
{
	dprintk("modem tty send xchar...\n");
}

static void modem_tty_throttle(struct tty_struct *tty)
{
	dprintk("modem tty throttle...\n");
}

static void modem_tty_unthrottle(struct tty_struct *tty)
{
	dprintk("modem tty unthrottle...\n");
}
#endif

/* write family */

static void modem_escape(unsigned long data)
{
	struct modem *m = (struct modem *)data;
	MODEM_DBG("do_modem_escape (count %d)...\n", m->escape_count);
        if (m->state == STATE_MODEM_ONLINE && m->escape_count == 3) {
                modem_set_state(m, STATE_COMMAND_ONLINE);
		if(m->tty) modem_tty_flush_buffer(m->tty);
		modem_put_chars(m,CRLF_CHARS(m),2);
		modem_report_result(m,STATUS_OK);  // fixme
        }
	m->escape_count = 0;
}

// FIXME: improve escape handling
static int modem_check_escape(struct modem *m, int from_user,
			      const u_char *buffer, int count)
{
	u_char esc_buf[3];
	unsigned long flags;
	const u_char *buf;
	int i; int is_escape = 1;
	if(count + m->escape_count > 3 ) {
		m->escape_count = 0;
		m->last_esc_check = jiffies;
		return 0;
	}
	if(from_user) {
		down(&m->sem);
		copy_from_user(esc_buf,buffer,count);
		buf = esc_buf;
	}
	else
		buf = buffer;

	for( i = 0 ; i < count ; i++) {
		if(buf[i] != ESCAPE_CHAR(m)) {
			is_escape = 0;
			break;
		}
	}
	spin_lock_irqsave(&m->lock,flags);
	if(!is_escape) {
		m->escape_count = 0;
		goto out;
	}
	if(m->escape_count == 0) {
		if(time_before(jiffies,m->last_esc_check + ESCAPE_TIMEOUT(m)))
			goto out;
	}
	else if (time_after(jiffies,m->last_esc_check + ESCAPE_TIMEOUT(m))) {
		m->escape_count = 0;
		goto out;
	}
	m->escape_count += count;
	if(m->escape_count == 3) {
		m->timer.expires  = jiffies + ESCAPE_TIMEOUT(m);
		m->timer.function = modem_escape;
		m->timer.data     = (unsigned long)m;
		add_timer(&m->timer);
	}
 out:
	m->last_esc_check = jiffies;
	spin_unlock_irqrestore(&m->lock,flags);
	if(from_user) {
		up(&m->sem);
	}
	return is_escape;
}

static int modem_tty_write(struct tty_struct *tty, int from_user,
			   const u_char *buffer, int count)
{
	struct modem *m = tty->driver_data;
	unsigned long flags;
       	const u_char *buf;
	int cnt, ret = 0;
	if (from_user && verify_area(VERIFY_READ,buffer,count))
		return -EFAULT;
	if(IS_STATE_CONNECTING(m->state)) {
		MODEM_DBG("modem_tty_write: hangup...\n");
		modem_hup(m,1);
		return count;
	}
	if(m->state == STATE_MODEM_ONLINE) // FIXME: also DISC
		modem_check_escape(m,from_user,buffer,count);
	if(from_user)
		down(&m->sem);
	while(count) {
		cnt = count;
		spin_lock_irqsave(&m->lock,flags);
		if(cnt > m->xmit.size - m->xmit.count)
			cnt = m->xmit.size - m->xmit.count;
		if(cnt > m->xmit.size - m->xmit.head)
			cnt = m->xmit.size - m->xmit.head;
		if(cnt <= 0) {
			spin_unlock_irqrestore(&m->lock,flags);
			break;
		}
		if(from_user) {
			spin_unlock_irqrestore(&m->lock,flags);
			if(cnt > sizeof(m->write_buf))
				cnt = sizeof(m->write_buf);
			copy_from_user(m->write_buf,buffer,cnt);
			buf = m->write_buf;
			spin_lock_irqsave(&m->lock,flags);
			if(cnt > m->xmit.size - m->xmit.count)
				cnt = m->xmit.size - m->xmit.count;
			if(cnt > m->xmit.size - m->xmit.head)
				cnt = m->xmit.size - m->xmit.head;
			if(cnt <= 0) {
				spin_unlock_irqrestore(&m->lock,flags);
				break;
			}
		}
		else
			buf = buffer;
                memcpy(m->xmit.buf + m->xmit.head, buf, cnt);
                m->xmit.count += cnt;
                m->xmit.head = (m->xmit.head + cnt)%m->xmit.size;
		spin_unlock_irqrestore(&m->lock,flags);
		ret += cnt;
		buffer += cnt;
		count -= cnt;
	}
	if(from_user)
		up(&m->sem);
	if(IS_STATE_COMMAND(m->state)) {
		queue_task(&m->at_tq,&tq_timer);
	}
	//MODEM_DBG("modem tty write: %d...\n",ret);
	return ret;
}

static int modem_tty_chars_in_buffer(struct tty_struct *tty)
{
	struct modem *m = tty->driver_data;
	//MODEM_DBG("modem tty chars_in_buffer: %d...\n",m->xmit.count);
	return m->xmit.count;
}

static void modem_tty_flush_buffer(struct tty_struct *tty)
{
	struct modem *m = tty->driver_data;
	unsigned long flags;
	MODEM_DBG("modem tty flush buffer...\n");
	spin_lock_irqsave(&m->lock,flags);
	m->xmit.head = m->xmit.tail = m->xmit.count = 0;
	if ((test_bit(TTY_DO_WRITE_WAKEUP, &tty->flags)) &&
	    tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
	wake_up_interruptible(&tty->write_wait);
	spin_unlock_irqrestore(&m->lock,flags);
}

static void modem_tty_break_ctl(struct tty_struct *tty, int state)
{
	struct modem *m = tty->driver_data;
	MODEM_DBG("tty break ctl...\n");
	//if (!MODEM_READY(m)) return;
	//modem_send_break(m);
}

static int modem_tty_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct modem *m = tty->driver_data;
	unsigned int info,old_info;
	//MODEM_DBG("modem tty ioctl: cmd %x, arg %lx...\n",cmd,arg);
	switch (cmd) {
	case TIOCMGET:
		info = m->modem_info;
		info &= TIOCM_DTR|TIOCM_RTS|TIOCM_DSR|TIOCM_RI|TIOCM_CD;
		info |= TIOCM_CTS;
		return put_user(info,(unsigned long *)arg);
	case TIOCMSET:
	case TIOCMBIS:
	case TIOCMBIC:
		if (get_user(info,(unsigned long *) arg))
			return -EFAULT;
		old_info = m->modem_info;
		info &= (TIOCM_DTR|TIOCM_RTS);
		switch (cmd) {
		case TIOCMSET:
			m->modem_info = (m->modem_info&~(TIOCM_DTR|TIOCM_RTS))|info;
			break;
		case TIOCMBIS:
			m->modem_info |= info;
			break;
		case TIOCMBIC:
			m->modem_info &= ~info;
			break;
		}
		// TODO: hup() then DTR is dropped
		// NOTE: In case of droped DTR disconnection 'NOCARRIER'
		// message should be sent to tty after DTR restoring
		// ...
		if(m->modem_info != old_info) {
			MODEM_DBG("ioctl: line ctrl %s|%s -> %s|%s\n",
				  old_info&TIOCM_DTR?"DTR":"",
				  old_info&TIOCM_RTS?"RTS":"",
				  m->modem_info&TIOCM_DTR?"DTR":"",
				  m->modem_info&TIOCM_RTS?"RTS":"");
		}
		return 0;
	case TIOCGSERIAL:
	case TIOCSSERIAL:
		/* virtual UART is protected by PCTel's patent */
		break;
	}
	return -ENOIOCTLCMD;
}

static void modem_tty_set_termios(struct tty_struct *tty, struct termios *old)
{
	struct modem *m = tty->driver_data;
	if (old && tty->termios->c_cflag == old->c_cflag)
		return;
	MODEM_DBG("tty set termios...\n");
        if(!tty_get_baud_rate(tty)) {
                MODEM_DBG("modem_tty_set_termios: hangup.\n");
		if(m->state!=STATE_MODEM_IDLE)
			modem_hup(m,1);
			// TBD: drop DTR?
        }
	else
		m->modem_info |= TIOCM_DTR;
}


#if 0
static void modem_tty_hangup(struct tty_struct *tty)
{
	struct modem *m = tty->driver_data;
	MODEM_DBG("modem_tty_hangup... %d\n", m->open_count);
	modem_shutdown(m);
	m->open_count = 0;
	m->tty = NULL;
}

static void modem_tty_stop(struct tty_struct *tty)
{
	dprintk("modem tty stop...\n");
}

static void modem_tty_start(struct tty_struct *tty)
{
	dprintk("modem tty start...\n");
}

static void modem_tty_wait_until_sent(struct tty_struct *tty, int timeout)
{
	dprintk("modem tty wait until sent...\n");
}
#endif

// FIXME: race conditions handling - use sem

static int modem_tty_open(struct tty_struct *tty, struct file *filp)
{
	struct modem *m;
	int minor, ret;

	minor = MINOR(tty->device);
	if (minor >= NUM_OF_MODEMS ||
	    (m = modem_table[minor]) == NULL)
		return -ENODEV;

	MODEM_DBG("modem tty open...\n");
	MOD_INC_USE_COUNT;	
	if (m->open_count++ == 0) {
		MODEM_DBG("startup modem...\n");
		m->tty = tty;
		tty->low_latency = 1; /* need with high data rates */
		m->xmit.buf = (unsigned char *)__get_free_page(GFP_KERNEL);
		if ( !m->xmit.buf ) {
			ret = -ENOMEM;
			goto error;
		}
		m->xmit.size = PAGE_SIZE;
		m->xmit.head = m->xmit.tail = m->xmit.count = 0;
		if(m->driver.open &&
		   (ret = m->driver.open(m)) != 0) {
			m->xmit.size = 0;
			m->xmit.count = m->xmit.tail = m->xmit.head = 0;
			free_page((unsigned long)m->xmit.buf);
			m->xmit.buf = 0;
			goto error;
		}
		*tty->termios = m->saved_termios;
		m->modem_info |= TIOCM_DTR|TIOCM_RTS;
		// TODO: update speed,DTR according to termios
		// ...
	}
	tty->driver_data = m;
	return 0;
 error:
	MODEM_ERR("cannot open modem %d.\n", minor);
	MOD_DEC_USE_COUNT;	
	return ret;
}

static void modem_tty_close(struct tty_struct *tty, struct file *filp)
{
	struct modem *m = tty->driver_data;
	if (!m)
		return ;
	MODEM_DBG("modem tty close...\n");
	if (--m->open_count <= 0) {
		MODEM_DBG("shutdown modem...\n");
		/* wait until send */
		// ... 
		if(C_HUPCL(tty)) {
			m->modem_info &= ~(TIOCM_DTR|TIOCM_RTS);
			modem_hup(m,1);
		}
		if(m->driver.close)
			m->driver.close(m);
		m->tty = 0;
		m->saved_termios = *tty->termios;
		m->xmit.size = 0;
		m->xmit.count = m->xmit.tail = m->xmit.head = 0;
		free_page((unsigned long)m->xmit.buf);
		m->xmit.buf = 0;
		m->open_count = 0;
	}
	MOD_DEC_USE_COUNT;
}

/* ---------------------------------------------------------------- */

#define MODEM_TTY_MAJOR 212 /* 212 */

static int modem_tty_refcount;
static struct tty_struct *modem_tty_table[NUM_OF_MODEMS];
static struct termios *modem_tty_termios[NUM_OF_MODEMS];
static struct termios *modem_tty_termios_locked[NUM_OF_MODEMS];



static struct tty_driver modem_tty_driver = {
        magic:                  TTY_DRIVER_MAGIC,
        driver_name:            "modem tty",
#ifdef CONFIG_DEVFS_FS
	name:                   "ttySL/%d",
#else
        name:                   "ttySL",
#endif
        major:                  MODEM_TTY_MAJOR,
        minor_start:            0,
        num:                    NUM_OF_MODEMS,
        type:                   TTY_DRIVER_TYPE_SERIAL,
        subtype:                SERIAL_TYPE_NORMAL,
        flags:                  TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS,

        refcount:               &modem_tty_refcount,
        table:                  modem_tty_table,
        termios:                modem_tty_termios,
        termios_locked:         modem_tty_termios_locked,

        open:                   modem_tty_open,
        close:                  modem_tty_close,
        write:                  modem_tty_write,
	//put_char:               modem_tty_put_char,
        //flush_chars:            modem_tty_flush_chars,
        write_room:             modem_tty_write_room,
        chars_in_buffer:        modem_tty_chars_in_buffer,
        ioctl:                  modem_tty_ioctl,
        set_termios:            modem_tty_set_termios,
        //throttle:               modem_tty_throttle,
        //unthrottle:             modem_tty_unthrottle,
	//stop:                   modem_tty_stop,
	//start:                  modem_tty_start,
	//hangup:                 modem_tty_hangup,
        break_ctl:              modem_tty_break_ctl,
	flush_buffer:           modem_tty_flush_buffer,
	//set_ldisc:NULL,
	//wait_until_sent:        modem_tty_wait_until_sent,
	//send_xchar:             modem_tty_send_xchar,
	//read_proc:NULL,write_proc:NULL,
};


/* ---------------------------------------------------------------- */

struct modem *modem_create(struct modem_driver *drv, char *dev_name)
{
	struct modem *m;
	int i;
	m = kmalloc(sizeof(*m), GFP_KERNEL);
	if (!m)
		return NULL;
	memset(m,0,sizeof(*m));

	spin_lock_init(&m->lock);
	init_MUTEX(&m->sem);

	m->device_name = dev_name;
	m->driver = *drv;
        m->modem_info = 0;
	m->state  = STATE_MODEM_IDLE;
        m->hook   = MODEM_HOOK_ON;
        m->caller = 0;
        m->capabilities = MODEM_CAPABILITIES;
	m->min_rate = MODEM_MIN_RATE;
	m->max_rate = MODEM_MAX_RATE;

	m->modem_info |= (TIOCM_DSR|TIOCM_CTS);

        sregs_init(m->sregs);

	if(modem_homolog_init(m,-1,default_country) &&
	   modem_homolog_init(m,MODEM_DEFAULT_COUNTRY_CODE,NULL)) {
		MODEM_ERR("bad country and country_code %x!\n",
			  MODEM_DEFAULT_COUNTRY_CODE);
		kfree(m);
		return NULL;
	}

	init_timer(&m->timer);
	INIT_TQUEUE(&m->start_task,do_modem_start,m);
	INIT_TQUEUE(&m->stop_task,do_modem_stop,m);
	INIT_TQUEUE(&m->change_dp_task,do_modem_change_dp,m);

	INIT_TQUEUE(&m->at_tq,modem_at_process,m);

	/* setup config */
	modem_setup_config(m);
	/* packer initializations */
	// fixme: TBD

        /* dp initialization */
        m->format = MODEM_FORMAT;
        m->srate  = MODEM_RATE;
        m->frag   = MODEM_FRAG; /* in bytes */
        m->dp = NULL;

	down(&modem_sem);
	for( i = 0 ; i < NUM_OF_MODEMS ; i++ )
		if (modem_table[i] == NULL) {
			modem_table[i] = m;
			m->minor = i;
			m->saved_termios = modem_tty_driver.init_termios;
			break;
		}
	up(&modem_sem);

	if (i==NUM_OF_MODEMS) {
		kfree(m);
		return NULL;
	}

	/* setup name */
	sprintf(m->name,"ttySL%d",i);

	printk(KERN_INFO "%s: %s\n", m->name,m->device_name);

	return m;
}


void modem_delete(struct modem *m)
{
	int i;
	MODEM_DBG("modem_delete...\n");
	if (m->tty) {
		/* FIXME: TBD */
		MODEM_DBG("opened modem.\n");
		tty_hangup(m->tty);
		flush_scheduled_tasks();
		if(m->driver.close)
			m->driver.close(m);
		m->tty->driver_data = NULL;
		m->tty              = 0;
		m->open_count       = 0;
		m->xmit.size = m->xmit.count = m->xmit.tail = m->xmit.head = 0;
		free_page((unsigned long)m->xmit.buf);
		m->xmit.buf = 0;
	}
	down(&modem_sem);
	for( i = 0 ; i < NUM_OF_MODEMS ; i++ )
		if (modem_table[i] == m) {
			modem_table[i] = NULL;
			break;
		}
	up(&modem_sem);
	kfree(m);
}


/* ---------------------------------------------------------------- */

EXPORT_SYMBOL_NOVERS(modem_create);
EXPORT_SYMBOL_NOVERS(modem_delete);
EXPORT_SYMBOL_NOVERS(modem_error);
EXPORT_SYMBOL_NOVERS(modem_ring);
EXPORT_SYMBOL_NOVERS(modem_process);

/* module parameters */

#ifdef CONFIG_DEBUG_MODEM
static int __initdata debug = 0;
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug level: 0-4, default=0");
extern int  modem_debug_init(int level);
extern void modem_debug_exit(void);
#else
#define modem_debug_init(level)
#define modem_debug_exit()
#endif

static char __initdata *country;
MODULE_PARM(country, "s");
MODULE_PARM_DESC(country, "Country name");



/* modem init externals : FIXME remove it */
extern int modem_init(void);
extern int modem_exit(void);


static int __init modem_tty_init(void)
{
	int i;

        for (i = 0; i < NUM_OF_MODEMS ; ++i) {
                modem_table[i] = NULL;
        }

	default_country = country;
	modem_debug_init(debug);
	modem_init();
	modem_tty_driver.init_termios = tty_std_termios;
	modem_tty_driver.init_termios.c_cflag = B9600|CS8|CREAD|HUPCL|CLOCAL;

        if (tty_register_driver(&modem_tty_driver)) {
		printk(KERN_ERR "cannot register modem tty driver\n");
                return -1;
	}

        printk(KERN_INFO "%s: version %s %s\n",
		modem_name,modem_version,modem_date);
	return 0;
}

static void __exit modem_tty_exit(void)
{
	tty_unregister_driver(&modem_tty_driver);
	modem_exit();
	modem_debug_exit();
}

module_init(modem_tty_init);
module_exit(modem_tty_exit);

MODULE_AUTHOR(MODEM_AUTHOR);
MODULE_DESCRIPTION(MODEM_NAME);
MODULE_LICENSE("Smart Link Ltd.");


